#
# ArchiveModel.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" .cati package archive handling """

import os
import mimetypes
import tarfile
import json
from .PackageJsonValidator import PackageJsonValidator
from .Pkg import Pkg
from . import PkgConvertor

class BaseArchive(Pkg):
    """ base archive for archive versions """
    def __init__(self, file_path: str, type_str: str):
        self.tar = tarfile.open(file_path, type_str)

    def add(self, path: str, arcname=None):
        """
        Add a file to package archive (in `w` mode)

        Args:
            path: that file you want to add
            arcname: the arcname
        """
        return self.tar.add(path, arcname=arcname)

    def close(self):
        """ Close package archive """
        return self.tar.close()

    def extractall(self, path: str):
        """ Extract all of package files to `path` arg """
        return self.tar.extractall(path)

    def pkg_version(self) -> str:
        """
        Returns dotcati package strcuture version

        Returns:
            '1.0' str
        """
        for member in self.tar.getmembers():
            if member.path == 'cati-version':
                # load dotcati package version
                f = self.tar.extractfile(member)
                return f.read().strip()
        # default version
        return '1.0'

    def members(self) -> list:
        """
        Returns members of the archive as paths string

        Returns:
            list[str] list of path strings
        """
        files = []
        for member in self.tar.getmembers():
            if member.path != '':
                files.append(member.path)
        return files

    def read(self):
        """
        Load package information on object.
        
        This method has not arg or output, just loads some information of package to this object.

        Raises:
            Exception: will raise normal Exception when package json data is invalid
                       (validated by dotcati.PackageJsonValidator.validate())
        """
        self.data = self.info()
        if not PackageJsonValidator.validate(self.data):
            raise
        # try to compare version for version validation
        Pkg.compare_version(self.data['version'], '0.0.0')
        self.data['version'] = self.data['version'].strip()

    def info(self) -> (dict, None):
        """ Returns package data.json information """
        for member in self.tar.getmembers():
            if member.path == 'data.json':
                f = self.tar.extractfile(member)
                return json.loads(f.read())
        return None

def archive_factory(file_path: str, type_str: str) -> BaseArchive:
    """
    Archive model factory. Loads a package archive from file

    the strcuture of packages, maybe change in new version of cati.
    so, cati should be compatible with old packages where
    created with old version of cati. this class
    is a factory to check package version and
    return archive model object by that
    version.

    Args:
        file_path: (str) archive filepath
        type_str: (str) open type of archive (r,w)

    Returns:
        returns instance of `dotcati.ArchiveModel.BaseArchive` as loaded package object
    """
    if os.path.isfile(file_path):
        if mimetypes.guess_type(file_path)[0] == 'application/x-debian-package':
            # package is a deb package
            # convert deb2cati
            file_path = PkgConvertor.deb2cati(file_path)
        elif mimetypes.guess_type(file_path)[0] == 'application/x-redhat-package-manager':
            # package is a rpm package
            # convert rpm2cati
            file_path = PkgConvertor.rpm2cati(file_path)

    # open v1 as default
    pkg = ArchiveModelV1(file_path, type_str)
    if pkg.pkg_version() == '1.0':
        return ArchiveModelV1(file_path, type_str)
    # return v1 object by default
    return ArchiveModelV1(file_path, type_str)

class ArchiveModelV1(BaseArchive):
    """ .cati package file model (v1.0) """
    pass



#
# File.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" repo file driver """

import os
import json
from cati.repo.drivers.BaseDriver import BaseDriver
from cati.repo.DirCrawler import DirCrawler

class File(BaseDriver):
    """ repo file driver """
    def test(self) -> bool:
        """ test repo """
        dir_path = self.url.split('://', 1)[-1]
        return os.path.isdir(dir_path) and os.access(dir_path, os.R_OK)

    def get_data(self, download_event=None) -> str:
        """ Returns repo data """
        crawler = DirCrawler(self.url.split('://', 1)[-1])
        crawler.start()
        try:
            return json.dumps(crawler.loaded_packages)
        except:
            return '[]'



#
# InstallCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Install command """

import os
from cati.cmdline.BaseCommand import BaseCommand
from cati.dotcati.Pkg import Pkg
from cati.frontend import RootRequired, Env
from cati.cmdline import pr, ansi, ArgParser
from cati.transaction.Calculator import Calculator
from cati.cmdline.components import TransactionShower
from .DownloadCommand import DownloadCommand
from .RemoveCommand import RemoveCommand
from .PkgCommand import PkgCommand
from cati.helpers.hash import calc_file_sha256, calc_file_md5

class InstallCommand(BaseCommand):
    """ Install command """
    def help(self):
        """
        install packages

        Usage: cati install [options] pkg1 pkg2 pkg3 ...
        ...... cati install [options] pkg1 pkg2=<version> ...

        Options:
        -y|--yes: don't ask for user confirmation
        --reinstall: reinstall gived packages
        --download-only: only download packages. this helps you to only download packages and install them later
        --with-recommends: also install recommended packages
        --keep-conffiles: `pkg install --keep-conffiles`
        --without-scripts: `remove/install --without-scripts`
        --target=[target-installation-path]: `pkg install --target`
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'install',
            'options': {
                '-y': [False, False],
                '--yes': [False, False],
                '--reinstall': [False, False],
                '--download-only': [False, False],
                '--with-recommends': [False, False],
                '--keep-conffiles': [False, False],
            },
            'max_args_count': None,
            'min_args_count': 1,
        }

    def set_manual_installs(self, packages):
        """ Sets installed packages type (manual/auto) """
        for pkg in packages:
            if not pkg.is_manual:
                path = Env.installed_lists('/' + pkg.data['name'] + '/manual')
                if os.path.isfile(path):
                    os.remove(path)

    def run(self):
        """ Run command """

        RootRequired.require_root_permission()

        pr.p('Loading packages list...')
        pr.p('========================')

        loaded_packages = []

        for argument in self.arguments:
            arg_parts = argument.split('=')
            if len(arg_parts) == 1:
                # load last version as default
                pkg = Pkg.load_last(argument)
            else:
                # load specify version
                pkg = Pkg.load_version(arg_parts[0], arg_parts[1])
                if pkg == 1:
                    pkg = False
                elif pkg == 2:
                    self.message('package "' + arg_parts[0] + '" has not version "' + arg_parts[1] + '"' + ansi.reset, before=ansi.red)
                    continue
            if pkg:
                loaded_packages.append(pkg)
            else:
                self.message('unknow package "' + argument + '"' + ansi.reset, before=ansi.red)

        # remove local packages from list
        new_loaded_packages = []
        for pkg in loaded_packages:
            try:
                file_path = pkg.data['file_path']
                new_loaded_packages.append(pkg)
            except:
                self.message('package "' + pkg.data['name'] + '" is a local package', is_error=True)
        loaded_packages = new_loaded_packages

        if not loaded_packages:
            return 1

        # calculate transactions
        pr.p('Calculating transactions...')
        calc = Calculator(with_recommends=self.has_option('--with-recommends'))
        i = 0
        while i < len(loaded_packages):
            loaded_packages[i].is_manual = True
            i += 1
        try:
            calc.install(list(reversed(loaded_packages)))
        except:
            pr.e(ansi.red + 'ERROR: There is some dependnecy problems.' + ansi.reset)
            return 1

        # handle reinstallable packages
        i = 0
        while i < len(calc.to_install):
            if calc.to_install[i].installed():
                if calc.to_install[i].installed() == calc.to_install[i].wanted_version:
                    if not self.has_option('--reinstall'):
                        pr.p('Package ' + calc.to_install[i].data['name'] + '=' + calc.to_install[i].wanted_version + ' is currently installed. use --reinstall option to re-install it.')
                        if calc.to_install[i].is_manual:
                            try:
                                pr.p('Setting it as manual installed package...')
                                manual_f = open(Env.installed_lists('/' + pkg.data['name'] + '/manual'), 'w')
                                manual_f.write('')
                                manual_f.close()
                            except:
                                pass
                        calc.to_install.pop(i)
            i += 1

        # check transactions exists
        if not calc.has_any_thing():
            pr.p('Nothing to do.')
            return 0

        # show transaction
        TransactionShower.show(calc)

        if not self.has_option('-y') or self.has_option('--yes'):
            pr.p('Do you want to continue? [Y/n] ', end='')
            answer = input()
            if answer == 'y' or answer == 'Y' or answer == '':
                pass
            else:
                pr.p('Abort.')
                return 1

        # start download packages
        pr.p('Downloading packages...')
        downloaed_paths = []
        for pkg in calc.to_install:
            download_path = Env.cache_dir('/archives/' + pkg.data['name'] + '-' + pkg.wanted_version + '-' + pkg.data['arch'])
            if os.path.isfile(download_path):
                file_sha256 = calc_file_sha256(download_path)
                file_md5 = calc_file_md5(download_path)
                valid_sha256 = None
                valid_md5 = None
                try:
                    valid_sha256 = pkg.data['file_sha256']
                except:
                    valid_sha256 = file_sha256
                try:
                    valid_md5 = pkg.data['file_md5']
                except:
                    valid_md5 = file_md5
                if file_md5 != valid_md5 or file_sha256 != valid_sha256:
                    # file is corrupt and should be download again
                    os.remove(download_path)
                else:
                    pr.p('Using Cache for ' + pkg.data['name'] + ':' + pkg.data['version'] + ':' + pkg.data['arch'] + '...')
                    downloaed_paths.append(download_path)
                    continue
            download_cmd = DownloadCommand()
            i = 0
            res = 1
            tmp = True
            while tmp:
                if i > 5:
                    pr.e(ansi.red + 'Failed to download packages' + ansi.reset)
                    return res
                pr.p('Downloading ' + pkg.data['name'] + ':' + pkg.data['version'] + ':' + pkg.data['arch'] + '...')
                res = download_cmd.handle(ArgParser.parse(['cati', 'download', '-q', pkg.data['name'] + '=' + pkg.wanted_version, '--output=' + download_path]))
                if res == 1 or res == None:
                    tmp = False
                i += 1
            downloaed_paths.append(download_path)
        pr.p('Download completed.')

        # check --download-only option
        if self.has_option('--download-only'):
            return 0

        cati_pkg_cmd_options = []
        remove_cmd_options = []
        if self.has_option('--keep-conffiles'):
            cati_pkg_cmd_options.append('--keep-conffiles')
        if self.has_option('--target'):
            cati_pkg_cmd_options.append('--target')
        if self.has_option('--without-scripts'):
            cati_pkg_cmd_options.append('--without-scripts')
            remove_cmd_options.append('--without-scripts')

        # remove packages
        if calc.to_remove:
            pr.p('Removing packages...')
            package_names = [pkg.data['name'] for pkg in calc.to_remove]
            remove_cmd = RemoveCommand()
            res = remove_cmd.handle(ArgParser.parse(['cati', 'remove', *package_names, '-y', *remove_cmd_options]))
            if res != 0 and res != None:
                pr.e(ansi.red + 'Failed to remove packages' + ansi.reset)
                return res

        # install packages
        pr.p('Installing packages...')
        pkg_cmd = PkgCommand()
        res = pkg_cmd.handle(ArgParser.parse(['cati', 'pkg', 'install', *downloaed_paths, *cati_pkg_cmd_options]))
        if res != 0 and res != None:
            self.set_manual_installs(calc.to_install)
            pr.e(ansi.red + 'Failed to install packages' + ansi.reset)
            return res

        self.set_manual_installs(calc.to_install)

        pr.p(ansi.green + 'Done.' + ansi.reset)



#
# ReposListErrorShower.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" shows repo config errors from list of repos """

from cati.cmdline import pr

def show(repos: list):
    """
    Shows errors in the repositories list

    Args:
        repos: list of loaded repositories (list[repo.Repo.Repo])
    """
    for repo in repos:
        if repo.syntax_errors:
            for error in repo.syntax_errors:
                pr.e(
                    'error in ' + repo.loaded_from_file + ':' + str(repo.line_number) + ': ' + error
                )



#
# ansi.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" ANSI terminal colors """

header = '\033[95m'
"""
header
"""

blue = '\033[94m'
"""
blue
"""

green = '\033[92m'
"""
green
"""

yellow = '\033[93m'
"""
yellow
"""

red = '\033[91m'
"""
red
"""

reset = '\033[0m'
"""
reset
"""

bold = '\033[1m'
"""
bold
"""

underline = '\033[4m'
"""
underline
"""

cyan = '\033[96m'
"""
cyan
"""

def disable():
    """
    disable ansi characters.
    assign null value to all of ansi chars to disable them
    """
    from cmdline import ansi
    ansi.header = ''
    ansi.blue = ''
    ansi.green = ''
    ansi.yellow = ''
    ansi.red = ''
    ansi.reset = ''
    ansi.bold = ''
    ansi.underline = ''
    ansi.cyan = ''



#
# Env.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

"""
Handle environment file paths.

in cati, we do not write config, database and... file paths directly in everywhere.
we get that paths from this module.
also testing system changes base environment path in `base_path_dir` variable
to isolate program environemnt while running tests
"""

base_path_dir = ''
"""
str: base path of the cati files environment.
(will get another path in testing environment to isolate testing environment with real environment)
"""

def base_path(path=''):
    """ Returns environment base path """
    return base_path_dir + path

def packages_lists(path=''):
    """ Packages list directory """
    return base_path('/var/lib/cati/lists' + path)

def installed_lists(path=''):
    """ Installed packages list directory """
    return base_path('/var/lib/cati/installed' + path)

def state_file():
    """ State file (read state system documentation) """
    return base_path('/var/lib/cati/state.f')

def unremoved_conffiles():
    """ Unremoved conffiles list filepath """
    return base_path('/var/lib/cati/unremoved-conffiles.list')

def security_blacklist(path=''):
    """ Security blacklist directory """
    return base_path('/var/lib/cati/security-blacklist' + path)

def any_scripts(path=''):
    """ Any scripts directory """
    return base_path('/var/lib/cati/any-scripts' + path)

def repos_config():
    return base_path('/etc/cati/repos.list')

def repos_config_dir(path=''):
    return base_path('/etc/cati/repos.list.d' + path)

def cache_dir(path=''):
    return base_path('/var/cache/cati' + path)

def allowed_archs():
    return base_path('/etc/cati/allowed-architectures.list')



#
# HealthChecker.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Checks cati installation health and repair """

import os
from . import Env

def repair_once_file(filepath: str, events: dict):
    """
    Repairs once file

    Args:
        filepath (str): the filepath to repair
        events (dict):
            - faild_to_repair
    """
    try:
        f = open(Env.base_path('/' + filepath).replace('//', '/'), 'w')
        f.write('')
        f.close()
    except:
        events['failed_to_repair']('/' + filepath, 'file')

def repair_once_dir(dirpath: str, events: dict):
    """
    Repairs once dir

    Args:
        dirpath (str): the dirpath to repair
        events (dict):
            - faild_to_repair
    """
    try:
        os.mkdir(Env.base_path('/' + dirpath))
    except:
        events['failed_to_repair']('/' + dirpath, 'dir')

def check(events: dict):
    """
    Check all of needed files and dirs for cati installation

    Args:
        events:
            - failed_to_repair: will run when cati installation is corrupt and user has not root permission
            to repair it and passes filepath and type of that to function
    """

    required_files = [
        '/var/lib/cati/state.f',
        '/var/lib/cati/unremoved-conffiles.list',
        '/etc/cati/repos.list',
        '/etc/cati/allowed-architectures.list',
    ]

    required_dirs = [
        '/var',
        '/var/lib',
        '/var/lib/cati',
        '/var/lib/cati/lists',
        '/var/lib/cati/installed',
        '/var/lib/cati/security-blacklist',
        '/var/lib/cati/any-scripts',
        '/var/cache',
        '/var/cache/cati',
        '/var/cache/cati/archives',
        '/etc',
        '/etc/cati',
        '/etc/cati/repos.list.d',
    ]

    for d in required_dirs:
        if not os.path.isdir(Env.base_path('/' + d)):
            repair_once_dir(d, events)

    for f in required_files:
        if not os.path.isfile(Env.base_path('/' + f)):
            repair_once_file(f, events)



#
# Scanner.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Scans an directory and create packages data files """

import os
import json
from cati.dotcati.ArchiveModel import archive_factory
from cati.helpers.hash import calc_file_sha256, calc_file_md5

def scan(directory: str):
    """
    Scans an directory and create packages data files
    
    Args:
        directory (str): the repository directory path
    """
    for item in os.listdir(directory):
        if os.path.isfile(directory + '/' + item):
            if item.split('.')[-1] in ['cati', 'deb', 'rpm']:
                try:
                    scan_once(directory + '/' + item)
                except:
                    pass
        else:
            scan(directory + '/' + item)

def scan_once(filepath: str):
    """
    Scan once package
    
    Args:
        filepath (str): the package filepath
    """
    pkg = archive_factory(filepath, 'r')
    pkg.read()
    data = pkg.data

    data['file_path'] = os.path.abspath(filepath)
    data['file_sha256'] = calc_file_sha256(filepath)
    data['file_md5'] = calc_file_md5(filepath)
    data['pkg_type'] = filepath.split('.')[-1]
    data['file_size'] = os.path.getsize(filepath)

    data_file = open(filepath + '.json', 'w')
    data_file.write(json.dumps(data))
    data_file.close()
    pkg.close()



#
# PackageShower.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

"""
Package shower cli component
"""

from cati.cmdline import ansi, pr
from cati.dotcati.Pkg import Pkg

def show(data: dict):
    """
    shows package information from data dictonary.
    
    Args:
        data: (dict) package data dictonary
    """
    output = ''
    output += 'Name: ' + ansi.green + data['name'] + ansi.reset + '\n'
    output += 'Version: ' + ansi.blue + data['version'] + ansi.reset + '\n'
    output += 'Arch: ' + ansi.yellow + data['arch'] + ansi.reset + '\n'
    try:
        author = data['author']
        output += 'Author: ' + ansi.header + author + ansi.reset + '\n'
    except:
        pass
    try:
        maintainer = data['maintainer']
        output += 'Maintainer: ' + ansi.cyan + maintainer + ansi.reset + '\n'
    except:
        pass
    try:
        changed_by = data['changed-by']
        output += 'Changed-By: ' + changed_by + '\n'
    except:
        pass
    try:
    	uploaders = data['uploaders']
    except:
    	uploaders = []
    if uploaders:
        output += 'Uploaders: '
        for uploader in uploaders:
            output += uploader + ', '
        output = output[:len(output)-2]
        output += '\n'
    try:
        channel = data['channel']
        output += 'Channel: ' + ansi.red + channel + ansi.reset + '\n'
    except:
        pass
    try:
        homepage = data['homepage']
        output += 'Homepage: ' + ansi.blue + homepage + ansi.reset + '\n'
    except:
        pass
    try:
        category = data['category']
        if category:
            output += 'Categories: '
            i = 0
            while i < len(category):
                output += ansi.bold + category[i] + ansi.reset
                if i < len(category)-1:
                    output += ', '
                i += 1
            output += '\n'
    except:
        pass
    try:
        description = data['description']
        output += 'Description: ' + description + '\n'
    except:
        pass
    try:
        changes = data['changes']
        output += 'Changes: ' + description + '\n'
    except:
        pass
    try:
        date = data['date']
        output += 'Last-Update-Date: ' + date + '\n'
    except:
        pass
    try:
        urgency = data['urgency']
        output += 'Update-Level: ' + urgency + '\n'
    except:
        pass
    try:
        essential = data['essential']
        if essential:
            essential = 'Yes'
        else:
            essential = 'No'
        output += 'Essential: ' + essential + '\n'
    except:
        pass
    try:
        file_size = Pkg.get_download_size_str(data['file_size'])
        output += 'Download-Size: ' + str(file_size) + '\n'
    except:
        pass
    try:
        installed_size = Pkg.get_download_size_str(data['installed-size'])
        output += 'Installed-Size: ' + str(installed_size) + '\n'
    except:
        pass
    try:
        repo = data['repo']
        output += 'Repository: ' + repo + '\n'
    except:
        pass
    try:
    	depends = data['depends']
    except:
    	depends = []
    if depends:
        output += 'Depends: '
        for dep in depends:
            output += ansi.bold + dep + ansi.reset + ', '
        output = output[:len(output)-2]
        output += '\n'
    try:
    	recommends = data['recommends']
    except:
    	recommends = []
    if recommends:
        output += 'Recommends: '
        for rec in recommends:
            output += ansi.bold + rec + ansi.reset + ', '
        output = output[:len(output)-2]
        output += '\n'
    try:
    	conflicts = data['conflicts']
    except:
    	conflicts = []
    if conflicts:
        output += 'Conflicts: '
        for conflict in conflicts:
            output += ansi.bold + conflict + ansi.reset + ', '
        output = output[:len(output)-2]
        output += '\n'
    try:
    	suggests = data['suggests']
    except:
    	suggests = []
    if suggests:
        output += 'Suggests: '
        for suggest in suggests:
            output += ansi.bold + suggest + ansi.reset + ', '
        output = output[:len(output)-2]
        output += '\n'
    try:
    	enhances = data['enhances']
    except:
    	enhances = []
    if enhances:
        output += 'Enhances: '
        for enhance in enhances:
            output += ansi.bold + enhance + ansi.reset + ', '
        output = output[:len(output)-2]
        output += '\n'
    try:
    	provides = data['provides']
    except:
    	provides = []
    if provides:
        output += 'Provides: '
        for provide in provides:
            output += ansi.bold + provide + ansi.reset + ', '
        output = output[:len(output)-2]
        output += '\n'
    if Pkg.is_installed(data['name']):
        installed_version = Pkg.installed_version(data['name'])
        if Pkg.is_installed_manual(data['name']):
            output += 'Installed-Manual: ' + installed_version + '\n'
        else:
            output += 'Installed: ' + installed_version + '\n'

    # show user defined fields
    for k in data:
        if k[0] == 'x' or k[0] == 'X':
            # that fields start with `x` character are user defined fields
            output += '\n' + k + ': ' + data[k]

    if output[-1] == '\n':
        output = output[:len(output)-1]
    pr.p(output.strip())



#
# Calculator.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

"""
Transaction calculator.

transaction calculator gets a list from packages for
install/remove/upgrade/downgrade operations
and calculates all of operations needed to be done
(actualy, includes dependencies, conflicts...)
"""

from cati.dotcati.Pkg import Pkg

class Calculator:
    """ Transaction calculator """
    def __init__(self, with_recommends=False):
        self.to_remove = []
        self.to_install = []
        self.with_recommends = with_recommends

    def has_any_thing(self):
        """ returns is there any transactions to do """
        return self.to_remove or self.to_install

    def get_total_download_size(self) -> int:
        """
        Returns total download size of packages

        Returns:
            int: the bytes count
        """

        total_size = 0

        for pkg in self.to_install:
            try:
                total_size += int(pkg.data['file_size'])
            except:
                pass

        return total_size

    def remove(self, pkgs: list):
        """ Add packages to remove """
        for pkg in pkgs:
            if pkg.installed():
                self.to_remove.append(pkg)
        self.refresh_lists()

    def get_sorted_list(self):
        """ returns sorted list of all of packages """
        new_list = []
        for item in self.to_remove:
            new_list.append(
                {
                    "action": "remove",
                    "pkg": item
                }
            )
        for item in self.to_install:
            new_list.append(
                {
                    "action": "install",
                    "pkg": item
                }
            )

        return new_list

    def install(self, pkgs: list):
        """ Add packages for install """
        for pkg in pkgs:
            try:
                tmp = pkg.wanted_version
            except:
                pkg.wanted_version = pkg.data['version']
            try:
                tmp = pkg.depend_get_next
            except:
                pkg.depend_get_next = {}
            try:
                tmp = pkg.conflict_get_next
            except:
                pkg.conflict_get_next = {}
            try:
                tmp = pkg.is_manual
            except:
                pkg.is_manual = False
            i = 0
            added = False
            while i < len(self.to_install):
                if self.to_install[i].data['name'] == pkg.wanted_version:
                    self.to_install[i] = pkg
                    added = True
                i += 1
            if not added:
                self.to_install.insert(0, pkg)
        self.refresh_lists()

    def handle_install_depends(self):
        """ Adds installable packages depends to install list """
        new_to_install = []
        i = 0
        while i < len(self.to_install):
            depends = self.to_install[i].get_depends()
            if self.with_recommends:
                depends = [*depends, *self.to_install[i].get_recommends()]
            for depend in depends:
                if not Pkg.check_state(depend) and depend.strip()[0] != '@':
                    # find package depend
                    try:
                        self.to_install[i].depend_get_next[depend]
                    except:
                        try:
                            self.to_install[i].depend_get_next[depend] = 0
                        except:
                            self.to_install[i].depend_get_next = {}
                            self.to_install[i].depend_get_next[depend] = 0
                    pkg = Pkg.check_state(depend, get_false_pkg=True, get_false_pkg_next=self.to_install[i].depend_get_next[depend])
                    self.to_install[i].depend_get_next[depend] += 1
                    if len(pkg) == 1:
                        a = 0
                        added = False
                        while a < len(self.to_install):
                            if self.to_install[a].data['name'] == pkg[0]:
                                added = True
                            a += 1
                        if not added:
                            new_to_install.append(Pkg.load_last(pkg[0]))
                    elif len(pkg) == 3:
                        a = 0
                        added = False
                        while a < len(self.to_install):
                            if self.to_install[a].data['name'] == pkg[0]:
                                wanted_version = pkg[2]
                                installed_version = self.to_install[a].wanted_version
                                if pkg[1] == '=':
                                    if Pkg.compare_version(installed_version, wanted_version) == 0:
                                        added = True
                                elif pkg[1] == '>=':
                                    if Pkg.compare_version(installed_version, wanted_version) >= 0:
                                        added = True
                                elif pkg[1] == '<=':
                                    if Pkg.compare_version(installed_version, wanted_version) <= 0:
                                        added = True
                                elif pkg[1] == '>':
                                    if Pkg.compare_version(installed_version, wanted_version) == 1:
                                        added = True
                                elif pkg[1] == '<':
                                    if Pkg.compare_version(installed_version, wanted_version) == -1:
                                        added = True
                            a += 1
                        if not added:
                            pkg_obj = None
                            if pkg[1] == '=':
                                pkg_obj = Pkg.load_version(pkg[0], pkg[2])
                            elif pkg[1] == '>=' or pkg[1] == '>':
                                pkg_obj = Pkg.load_last(pkg[0])
                            elif pkg[1] == '<=':
                                pkg_obj = Pkg.load_version(pkg[0], pkg[2])
                            elif pkg[1] == '<':
                                pkg_obj = Pkg.load_last(pkg[0])
                                versions = pkg_obj.get_versions_list()
                                x = 0
                                while x < len(versions):
                                    if Pkg.compare_version(versions[x][0], pkg[0]) >= 0:
                                        versions.pop(x)
                                    x += 1
                                versions = [v[0] for v in versions]
                                wanted_ver = pkg.get_last_version(versions)
                                pkg_obj = Pkg.load_version(pkg[0], wanted_ver)
                            new_to_install.append(pkg_obj)
            i += 1

        if new_to_install:
            self.install(new_to_install)

        self.handle_install_reverse_depends()

    def handle_install_reverse_depends(self):
        """ Adds installable packages reverse depends to install list """
        new_to_remove = []
        i = 0
        while i < len(self.to_install):
            reverse_depends = self.to_install[i].get_reverse_depends()
            for dep in reverse_depends:
                if dep.installed():
                    dep = Pkg.load_version(dep.data['name'], dep.installed())
                    for tmp_dep in dep.get_depends():
                        result = Pkg.check_state(tmp_dep, virtual={
                            'install': [
                                [
                                    self.to_install[i].data['name'], self.to_install[i].data['version']
                                ]
                            ],
                        })
                        if not result:
                            a = 0
                            added = False
                            while a < len(self.to_remove):
                                if self.to_remove[a].data['name'] == dep.data['name']:
                                    added = True
                                a += 1
                            if not added:
                                new_to_remove.append(dep)
            i += 1

        if new_to_remove:
            self.remove(new_to_remove)

    def handle_install_reverse_conflicts(self):
        """ Adds installable packages reverse conflicts to install list """
        new_to_remove = []
        i = 0
        while i < len(self.to_install):
            conflicts = self.to_install[i].get_reverse_conflicts()
            for conflict in conflicts:
                if conflict.installed():
                    a = 0
                    added = False
                    while a < len(self.to_remove):
                        if self.to_remove[a].data['name'] == conflict.data['name']:
                            added = True
                        a += 1
                    if not added:
                        new_to_remove.append(conflict)
            i += 1

        if new_to_remove:
            self.remove(new_to_remove)

    def handle_install_conflicts(self):
        """ Adds installable packages conflicts to install list """
        new_to_remove = []
        i = 0
        while i < len(self.to_install):
            conflicts = self.to_install[i].get_conflicts()
            for conflict in conflicts:
                if Pkg.check_state(conflict):
                    # find package conflict
                    try:
                        self.to_install[i].conflict_get_next[conflict]
                    except:
                        try:
                            self.to_install[i].conflict_get_next[conflict] = 0
                        except:
                            self.to_install[i].conflict_get_next = {}
                            self.to_install[i].conflict_get_next[conflict] = 0
                    pkg = Pkg.check_state(conflict, get_true_pkg=True, get_true_pkg_next=self.to_install[i].conflict_get_next[conflict])
                    self.to_install[i].conflict_get_next[conflict] += 1
                    a = 0
                    added = False
                    if type(pkg) == list:
                        while a < len(self.to_remove):
                            if self.to_remove[a].data['name'] == pkg[0]:
                                added = True
                            a += 1
                        if not added:
                            new_to_remove.append(Pkg.load_last(pkg[0]))
            i += 1

        if new_to_remove:
            self.remove(new_to_remove)
        
        self.handle_install_reverse_conflicts()

    def refresh_lists(self):
        """ Refresh packages list and sync them with depends and conflicts """
        # sync versions
        i = 0
        while i < len(self.to_remove):
            self.to_remove[i] = Pkg.load_version(self.to_remove[i].data['name'], self.to_remove[i].installed())
            i += 1
        i = 0
        while i < len(self.to_install):
            wv = self.to_install[i].wanted_version
            im = self.to_install[i].is_manual
            self.to_install[i] = Pkg.load_version(self.to_install[i].data['name'], self.to_install[i].wanted_version)
            self.to_install[i].wanted_version = wv
            self.to_install[i].is_manual = im
            i += 1

        # remove repeated lists in to_remove list
        a = 0
        while a < len(self.to_remove):
            b = 0
            while b < len(self.to_remove):
                if a != b:
                    if self.to_remove[a].data['name'] == self.to_remove[b].data['name']:
                        self.to_remove.pop(b)
                        a -= 1
                        b -= 1
                b += 1
            a += 1

        # sync to_remove list
        new_to_remove = []
        for item in self.to_remove:
            # load reverse dependnecy of current package and add them to list
            reverse_depends = item.get_reverse_depends()
            for rd in reverse_depends:
                if rd.installed():
                    if not rd.data['name'] in [tmp.data['name'] for tmp in self.to_remove]:
                        new_to_remove.append(rd)
        if new_to_remove:
            self.remove(new_to_remove)

        self.handle_install_depends()
        self.handle_install_conflicts()



#
# PackageJsonValidator.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Package information json validator """

class PackageJsonValidator:
    """ Package information json validator """
    @staticmethod
    def validate(data: dict) -> bool:
        """
        This function gets a json object and checks that fields and value of them where
        are required for a valid package data.json and says this data
        is valid or not

        Args:
            data: package data.json as dictonary

        Returns:
            boolean. True means data is valid and False means data is invalid
        """

        try:
            assert type(data['name']) == str
            assert type(data['version']) == str
            assert type(data['arch']) == str

            try:
                tmp = data['depends']
                try:
                    assert type(data['depends']) == list
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['conflicts']
                try:
                    assert type(data['conflicts']) == list
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['suggests']
                try:
                    assert type(data['suggests']) == list
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['conffiles']
                try:
                    assert type(data['conffiles']) == list
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['staticfiles']
                try:
                    assert type(data['staticfiles']) == list
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['description']
                try:
                    assert type(data['description']) == str
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['author']
                try:
                    assert type(data['author']) == str
                    assert not '|' in data['author']
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['maintainer']
                try:
                    assert type(data['maintainer']) == str
                    assert not '|' in data['maintainer']
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['homepage']
                try:
                    assert type(data['homepage']) == str
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['category']
                try:
                    assert type(data['category']) == list
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['channel']
                try:
                    assert type(data['channel']) == str
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['uploaders']
                try:
                    assert type(data['uploaders']) == list
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['changed-by']
                try:
                    assert type(data['changed-by']) == str
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['changes']
                try:
                    assert type(data['changes']) == str
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                tmp = data['date']
                try:
                    assert type(data['date']) == str
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                urgency = data['urgency']
                try:
                    assert type(data['urgency']) == str
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                installed_size = data['installed-size']
                try:
                    assert type(data['installed-size']) == int
                except:
                    raise
            except KeyError as ex:
                pass

            try:
                essential = data['essential']
                try:
                    assert type(data['essential']) == bool
                except:
                    raise
            except KeyError as ex:
                pass

            return True
        except:
            return False



#
# BaseCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Cmdline command model base """

import sys
from . import pr
from cati.frontend.Version import version as cati_version
from .commands import HelpCommand
from . import ArgParser

class BaseCommand:
    """ Cmdline command model base """

    def validate(self, args: dict):
        """
        Validate inserted arguments in command config frame
        loads command config from `config` function output
        next checks arguments and compares them with command config
        then, if an unknow option is inserted or more/less argument inserted,
        shows error to user.

        Args:
            args: command arguments as doctonary
                  example:
                  {'arguments': list, 'options': dict{ '--foo': 'value', '-x': None } }
        """

        # load command config
        command_config = self.config()
        # add --help option as default
        command_config['options']['--help'] = [False, False]

        self.name = command_config['name']
        self.cati_exec = sys.argv[0]

        # pop first argument (command self name)
        args['arguments'].pop(0)

        # check knowed options and value of them
        for k in args['options']:
            try:
                option_config = command_config['options'][k]
            except:
                self.message('unknow option "' + k + '"')
                return pr.exit(1)

            if option_config[1] == True:
                if args['options'][k] == None:
                    self.message('option ' + k + ' requires value')
                    return pr.exit(1)

        # check required options
        for option in command_config['options']:
            if command_config['options'][option][0] == True:
                try:
                    args['options'][option]
                except:
                    self.message('option ' + option + ' is required')
                    return pr.exit(1)

        self.args = args
        self.arguments = self.args['arguments']

        # check arguments count
        if not self.has_option('--help'):
            if command_config['max_args_count'] != None:
                if len(args['arguments']) > command_config['max_args_count']:
                    self.message('this command requires less than ' + str(command_config['max_args_count']+1) + ' arguments')
                    return pr.exit(1)

            if command_config['min_args_count'] != None:
                if len(args['arguments']) < command_config['min_args_count']:
                    self.message('this command requires more than ' + str(command_config['min_args_count']-1) + ' arguments')
                    return pr.exit(1)

    def handle(self, args: dict) -> (int, None):
        """
        Handle run the command
        first, validates arguments
        next, checks if --help inserted, show command help
        if not, run command

        Args:
            args: commands as dictonary
                  example:
                  {'arguments': list, 'options': dict{ '--foo': 'value', '-x': None } }
                  (this can be generated by )

        Returns:
            returns command exit code as int, or None value (None means 0 exit code)
        """
        self.validate(args)

        # handle --help option
        if self.has_option('--help'):
            help_cmd = HelpCommand.HelpCommand()
            return help_cmd.handle(ArgParser.parse(['cati', 'help', self.name]))

        return self.run()

    def has_option(self, option: str) -> bool:
        """
        Checks the option is inserted

        Args:
            option: (str) that option you want to check (for example `--foo`)

        Returns:
            boolean: if True, means option inserted, if False, means option is not inserted
        """
        try:
            self.args['options'][option]
            return True
        except:
            return False

    def option_value(self, option: str) -> (str, None):
        """
        Returns value of option

        Args:
            option: that option you want to get value of that
        
        Returns:
            `str` as value of option, if option has not value or not inserted, `None` will be returned

        Example:
            if `--foo=bar` command line option is inserted, if you pass `--foo` as `option`,
            you will get `bar` as output. if option has not assigned value, `None` will be returned
        """
        if not self.has_option(option):
            # if option not inserted, return None as default
            return None

        return self.args['options'][option]

    def message(self, msg, is_error=False, before=''):
        """
        Prints a message like this:
        `cati: <command-name>: <the-message>`

        Args:
            is_error: (bool) if this is True, message will print on stderr
            before: (str) before will print in the first of message
                    for example `message("hello world")` will print `cati: cmdname: hello world`
                    but `message("hello world", before="AAA ")` will print `AAA cati: cmdname: hello world`
        """
        msg = before + self.cati_exec + ': ' + self.name + ': ' + msg

        if is_error:
            pr.e(msg)
        else:
            pr.p(msg)

    def general_help(self) -> str:
        """
        returns general help of cati cli

        Returns: (str)
        """
        return """Cati package manager [""" + cati_version + """]
Copyright 2020-2021 parsa mpsh - GPL-3
Usage: cati [command] [options] [args]"""

    def help_full(self, with_general_help=True) -> str:
        """
        Returns full help of command
        the `with_general_help` argument used to include/exclude general help of cati

        Args:
            with_general_help: (bool) if is True, general help of cati will be included in output

        Returns:
            string help
        """
        help_text = self.help.__doc__
        help_text = help_text.strip()
        help_text_tmp = ''
        for line in help_text.split('\n'):
            line = line.strip()
            help_text_tmp += line + '\n'

        help_text = help_text_tmp.strip()

        if with_general_help:
            help_text = self.general_help() + '\n\n' + help_text

        return help_text

    def help_summary(self) -> str:
        """ Returns summary of help (first line only) """
        return self.help_full(False).split('\n')[0]

    def is_quiet(self) -> bool:
        """
        Checks --quiet and -q options

        Returns: boolean
        """
        return self.has_option('-q') or self.has_option('--quiet')

    def is_verbose(self) -> bool:
        """
        Checks --verbose and -v options 

        Returns: boolean
        """
        return self.has_option('-v') or self.has_option('--verbose')



#
# HelpCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Help command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import kernel, ansi, pr
from cati.frontend.Version import version as cati_version

class HelpCommand(BaseCommand):
    """ Help command """
    def help(self):
        """
        shows this help
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'help',
            'options': {
                '-v': [False, False],
                '--version': [False, False],
                '--quiet': [False, False],
                '-q': [False, False],
            },
            'max_args_count': 1,
            'min_args_count': 0,
        }

    def run(self):
        """ Run command """

        # check -v|--version options
        if self.has_option('--version') or self.has_option('-v'):
            pr.p(cati_version)
            return

        if not self.is_quiet():
            pr.p()
            pr.p(ansi.yellow + "  $$$$$$                $$     $$" + ansi.reset)
            pr.p(ansi.yellow + " $$    $$              $$" + ansi.reset)
            pr.p(ansi.yellow + " $$         $$$$$$   $$$$$$    $$" + ansi.reset)
            pr.p(ansi.yellow + " $$              $$    $$      $$" + ansi.reset)
            pr.p(ansi.yellow + " $$         $$$$$$$    $$      $$" + ansi.reset + ansi.cyan + " | " + self.general_help().split('\n')[0] + ansi.reset)
            pr.p(ansi.yellow + " $$    $$  $$    $$    $$  $$  $$" + ansi.reset + ansi.cyan + " | " + self.general_help().split('\n')[1] + ansi.reset)
            pr.p(ansi.yellow + "  $$$$$$    $$$$$$$     $$$$   $$" + ansi.reset + ansi.cyan + " | " + self.general_help().split('\n')[2] + ansi.reset)
            pr.p()

        pr.p('Options:')
        pr.p(ansi.header + '\t-v|--version' + ansi.reset + ': shows cati version')
        pr.p(ansi.header + '\t--no-ansi' + ansi.reset + ': disable terminal ansi colors')
        pr.p(ansi.header + '\t--help' + ansi.reset + ': pass it to commands to show help of that command')
        pr.p(ansi.header + '\t-q|--quiet' + ansi.reset + ': quiet output')
        pr.p(ansi.header + '\t-v|--verbose' + ansi.reset + ': verbose output')

        pr.p()

        commands = kernel.commands

        # check arguments
        if len(self.arguments) >= 1:
            # a command inserted. show detail of that command
            try:
                cmd = commands[self.arguments[0]]
                obj = cmd()
                pr.p(obj.help_full(False))
                return 0
            except:
                self.message('unknow command "' + self.arguments[0] + '"' + ansi.reset, True, before=ansi.red)
                return 1

        # show summary help of command
        pr.p('Subcommands:')
        for cmd in commands:
            obj = commands[cmd]()
            tabs = '\t\t'
            if len(cmd) > 7:
                tabs = '\t'
            pr.p('\t' + ansi.green + cmd + ansi.reset + tabs + obj.help_summary())

        if not self.is_quiet():
            pr.p('\nfor see detailed help about commands, run: "' + ansi.yellow + self.cati_exec + ' help [command-name]' + ansi.reset + '"')

        return 0



#
# AutoremoveCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Autoremove command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi, ArgParser
from cati.dotcati.Pkg import Pkg
from .RemoveCommand import RemoveCommand

class AutoremoveCommand(BaseCommand):
    """ Autoremove command """
    def help(self):
        """
        remove unused automaticaly installed packages

        Usage: sudo cati autoremove [options]

        Options: all of remove command options
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        rmcmd_options = RemoveCommand().config()['options']
        return {
            'name': 'autoremove',
            'options': rmcmd_options,
            'max_args_count': 0,
            'min_args_count': 0,
        }

    def find_unused_packages(self):
        """ finds unused packages """
        all_packages = Pkg.all_list()['list']

        for pkg in all_packages:
            try:
                if Pkg.is_installed(pkg.data['name']):
                    if not Pkg.is_installed_manual(pkg.data['name']):
                        has_any_reverse_depends = False
                        reverse_depends = Pkg.load_version(pkg.data['name'], Pkg.installed_version(pkg.data['name'])).get_reverse_depends()
                        for rdep in reverse_depends:
                            if rdep.installed():
                                rdep_is_in_unused_packages = False
                                for upkg in self.unused_packages:
                                    if upkg.data['name'] == rdep.data['name']:
                                        rdep_is_in_unused_packages = True
                                if not rdep_is_in_unused_packages:
                                    has_any_reverse_depends = True
                        if not has_any_reverse_depends:
                            self.unused_packages.append(pkg)
            except:
                pass

    def run(self):
        """ Run command """

        if not self.is_quiet():
            pr.p('Checking unused packages...')

        self.unused_packages = []
        self.find_unused_packages()
        unused_packages = self.unused_packages

        package_names = [pkg.data['name'] for pkg in unused_packages]
        options = [op for op in self.args['options']]

        if not package_names:
            pr.p('There is not any unused package')
            return 0

        remove_cmd = RemoveCommand()
        return remove_cmd.handle(ArgParser.parse(['cati', 'remove', *package_names, *options]))



#
# kernel.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Kernel of cli handler """

import sys
from . import ArgParser, pr
from .commands.HelpCommand import HelpCommand
from .commands.PkgCommand import PkgCommand
from .commands.ListCommand import ListCommand
from .commands.RemoveCommand import RemoveCommand
from .commands.ShowCommand import ShowCommand
from .commands.StateCommand import StateCommand
from .commands.QueryCommand import QueryCommand
from .commands.SearchCommand import SearchCommand
from .commands.FilesCommand import FilesCommand
from .commands.FinfoCommand import FinfoCommand
from .commands.RDependsCommand import RDependsCommand
from .commands.ForgetCommand import ForgetCommand
from .commands.CheckCommand import CheckCommand
from .commands.RepoCommand import RepoCommand
from .commands.UpdateCommand import UpdateCommand
from .commands.AutoremoveCommand import AutoremoveCommand
from .commands.ClearCacheCommand import ClearCacheCommand
from .commands.DownloadCommand import DownloadCommand
from .commands.InstallCommand import InstallCommand
from .commands.UpgradeCommand import UpgradeCommand
from .commands.FullUpgradeCommand import FullUpgradeCommand
from .commands.MeowCommand import MeowCommand

commands = {
    'help': HelpCommand,
    'pkg': PkgCommand,
    'list': ListCommand,
    'remove': RemoveCommand,
    'show': ShowCommand,
    'state': StateCommand,
    'query': QueryCommand,
    'search': SearchCommand,
    'files': FilesCommand,
    'finfo': FinfoCommand,
    'rdepends': RDependsCommand,
    'forget': ForgetCommand,
    'check': CheckCommand,
    'repo': RepoCommand,
    'update': UpdateCommand,
    'autoremove': AutoremoveCommand,
    'clear-cache': ClearCacheCommand,
    'download': DownloadCommand,
    'install': InstallCommand,
    'upgrade': UpgradeCommand,
    'full-upgrade': FullUpgradeCommand,
    'meow': MeowCommand,
}
"""
a dictonary from list of subcommands.
structure: "cmdname": CmdClass
"""

def handle(argv: list):
    """
    handle cli
    gets argv and runs entered command as subcommand (if not subcommand inserted, runs help command as default)

    Args:
        argv: the program arguments as list
    """
    # parse inserted arguments
    parsed_args = ArgParser.parse(argv)

    # find called subcommand by args
    if len(parsed_args['arguments']) == 0:
        # no subcommand called
        # call help command as default
        parsed_args['arguments'].append('help')

    try:
        command = commands[parsed_args['arguments'][0]]
    except:
        pr.e(sys.argv[0] + ': unknow command "' + parsed_args['arguments'][0] + '"')
        pr.exit(1)

    cmdobj = command()
    sys.exit(cmdobj.handle(parsed_args))



import setuptools

f = open('README.md', 'r')
long_description = f.read()
f.close()

setuptools.setup(
    name="cati",
    version="0.1-alpha5",
    author="parsa shahmaleki",
    author_email="parsampsh@gmail.com",
    description="The Cati Unix Package manager",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/parsampsh/gameoflife",
    packages=setuptools.find_packages(),
    scripts=['etc/bin/cati'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
    ],
    python_requires='>=3.6',
    install_requires=[
        'packaging >= 20.4',
        'requests >= 2.25.0',
        'wget >= 3.2',
    ],
)



#
# Http.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" repo http driver """

import os
import requests
from cati.repo.drivers.BaseDriver import BaseDriver
from cati.frontend import Temp

class Http(BaseDriver):
    """ repo http driver """
    def test(self) -> bool:
        """ test repo """
        try:
            res = requests.get(self.url + '?cati-repo-test=1')
        except:
            return False
        return res.ok

    def get_data(self, download_event=None) -> str:
        """ Returns repo data """
        i = 0
        last_res = None
        temp_file = Temp.make_file()
        os.remove(temp_file)
        while i < 5:
            if i > 4:
                return last_res
            last_res = download_event(self.url + '?get_data=1', temp_file)
            if last_res == True:
                f = open(temp_file, 'r')
                content = f.read().strip()
                f.close()
                if content != '':
                    return content
            i += 1



#
# SysArch.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Handle system architecture """

import os
from . import Env

is_testing = False
"""
in testing environment,
default architecture is `i386`, not real host arch.
"""

def sys_arch() -> str:
    """ Returns system architecture """
    if is_testing:
        return 'i386'
    else:
        arch = os.uname().machine
        if arch == 'i686':
            arch = 'i386'
        return arch

def allowed_archs() -> list:
    """
    Returns list of allowed architectures to install on the system

    The default archs are `all` and system architecture. also more archs
    will be loaded from /etc/cati/allowed-architectures.list

    Returns:
        list: list of allowed archs like ['all', 'amd64', '...']
    """
    # set default archs
    archs = ['all', sys_arch()]

    # load added archs
    f = open(Env.allowed_archs(), 'r')
    content = f.read()
    f.close()

    content = content.strip().split('\n')
    content = [line.strip() for line in content if line.strip() != '']
    archs = [*archs, *content]

    # remove repeated items from list
    archs = list(dict.fromkeys(archs))

    return archs



#
# RootRequired.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" A tool to check program permission and if it haven't root permission, die the program """

import os
import sys
from cati.cmdline import pr, ansi
from . import Env

is_testing = False
"""
bool: this variable will be True in testing environment to disable root permission check
(root permission in not needed in testing environment)
"""

def require_root_permission(is_cli=True, die_action=None):
    """
    checks root premission.

    Args:
        is_cli (bool): if is True, when user have not root permission,
            error will print in terminal. but if is False,
            the `die_action` will run as a function.
            (will be disable in testing environment)
        die_action (callable): the function will be run when `is_cli` is False
    """

    # if program is in testing mode don't check permission
    if is_testing:
        return

    if os.getuid() == 0:
        return

    # check write and read access for needed files
    files_to_check = [
        Env.packages_lists(),
        Env.installed_lists(),
        Env.state_file(),
        Env.unremoved_conffiles(),
        Env.security_blacklist(),
        Env.any_scripts(),
        Env.repos_config(),
        Env.repos_config_dir(),
        Env.cache_dir(),
        Env.allowed_archs(),
    ]
    for f in files_to_check:
        if not os.access(f, os.W_OK) or not os.access(f, os.R_OK):
            if is_cli:
                pr.e(ansi.red + sys.argv[0] + ': permission is denied' + ansi.reset)
                pr.exit(1)
                return
            else:
                die_action()
                return



#
# Builder.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Cati package builder """

import os
import json
import tarfile
from .exceptions import InvalidPackageDirException
from .ArchiveModel import archive_factory
from .PackageJsonValidator import PackageJsonValidator
from cati.frontend import Temp

class Builder:
    """ Cati package builder """

    @staticmethod
    def json_fields_are_valid(data: dict) -> bool:
        """
        alias for PackageJsonValidator.validate (validates package json info)

        Args:
            data: package data.json as dict
        
        Returns:
            returns boolean. True means json data is valid and False means invalid
        """
        return PackageJsonValidator.validate(data)

    def build(self, dirpath: str, output=None):
        """
        This function gets a directory and builds a .cati package from that.
        
        Args:
            dirpath: (str) package directory
            output: (str or None) path of the package output

        Raises:
            FileNotFoundError: when the directory not exists
            dotcati.exceptions.InvalidPackageDirException: when there is an probelm in package directory (problem will put as exception message)
        """

        if not os.path.isdir(dirpath):
            raise FileNotFoundError

        if not os.path.isfile(dirpath + '/data.json'):
            raise InvalidPackageDirException('file data.json not found')

        # validate data.json content
        try:
            data_json_content = open(dirpath + '/data.json', 'r').read()
        except:
            raise InvalidPackageDirException('cannot read data.json')

        try:
            data_json = json.loads(data_json_content)
        except:
            raise InvalidPackageDirException('invalid json syntax in data.json')

        if not self.json_fields_are_valid(data_json):
            raise InvalidPackageDirException('one or more fields in data.json is invalid')

        if not os.path.isdir(dirpath + '/files'):
            os.mkdir(dirpath + '/files')

        # compress and build package
        if output == None:
            # set default output file
            output = ''
            if dirpath[-1] == '/':
                output = dirpath[:-1]
            else:
                output = dirpath

            output += '.cati'

        self.compress(dirpath, output)

        return output

    def compress(self, dirpath: str, output: str):
        """
        This function compresses content of target directory and build package
        in output file

        Args:
            dirpath: (str) package directory path
            output: (str) that filepath you want to compress file in that

        Raises:
            InvalidPackageDirException: when the output path is invalid
        """
        try:
            pkg = archive_factory(output, "w:gz")
        except:
            raise InvalidPackageDirException('file "' + output + '" for output of package not found')

        pkg.add(dirpath, arcname='/')
        # add `cati-version` file
        cati_version_tmp_f = Temp.make_file()
        tmp_f = open(cati_version_tmp_f, 'w')
        tmp_f.write('1.0')
        tmp_f.close()
        pkg.add(cati_version_tmp_f, arcname='/')
        pkg.close()



#
# FullUpgradeCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" FullUpgrade command """

from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import ArgParser
from .AutoremoveCommand import AutoremoveCommand
from .UpdateCommand import UpdateCommand
from .UpgradeCommand import UpgradeCommand

class FullUpgradeCommand(BaseCommand):
    """ FullUpgrade command """
    def help(self):
        """
        update, upgrade and autoremove

        Usage: cati full-upgrade [options]

        Options:
        -y|--yes: don't ask for user confirmation
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'full-upgrade',
            'options': {
                '-y': [False, False],
                '--yes': [False, False],
            },
            'max_args_count': 0,
            'min_args_count': 0,
        }

    def run(self):
        """ Run command """

        options = []
        if self.has_option('-y') or self.has_option('--yes'):
            options = ['-y']

        update_cmd = UpdateCommand()
        res = update_cmd.handle(ArgParser.parse(['cati', 'update']))
        if res != 0 and res != None:
            return res

        upgrade_cmd = UpgradeCommand()
        res = upgrade_cmd.handle(ArgParser.parse(['cati', 'upgrade', *options]))
        if res != 0 and res != None:
            return res

        autoremove_cmd = AutoremoveCommand()
        res = autoremove_cmd.handle(ArgParser.parse(['cati', 'autoremove', *options]))
        if res != 0 and res != None:
            return res



#
# DownloadProgress.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Cli download progress bar """

import wget

def download(url: str, output_path=None) -> (bool, Exception):
    """
    Download `url` and save in `output_path` and shows progress in terminal

    Args:
        url: (str) that url you want to download
        output_path: (str) that path you want to download file in that (optional)

    Returns:
        bool True: means download is successful
        instance of Exception: means download faild and returns exception
    """
    try:
        wget.download(url, out=output_path)
        print()
        return True
    except Exception as ex:
        return ex



#
# BaseDriver.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Repo base driver """

class BaseDriver:
    """ Repo base driver """
    def __init__(self, url: str, pkg: str, arch: str):
        self.url = url
        self.pkg = pkg
        self.arch = arch



#
# pr.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Print handling in cli """

import sys
from cati.frontend import Temp

is_testing = False
"""
if this variable is True, print will be disable. this variable will get True value while running tests to hide prints
"""

def p(value='', end='\n'):
    """
    Prints on stdout

    Args:
        value: that value you want to print
        end: the end of line (default is `\\n`)
    """
    if is_testing:
    	return
    return print(value, end=end, flush=True)

def e(value='', end='\n'):
    """
    Prints on stderr

    Args:
        value: that value you want to print
        end: the end of line (default is `\\n`)
    """
    if is_testing:
    	return
    return print(value, end=end, flush=True, file=sys.stderr)

def exit(code=0):
    """
    Exits program with exit code and deletes temp files before exit

    Args:
        code: the exit code (default 0)
    """
    if is_testing:
    	return
    # delete temp files before exit
    Temp.clean()
    # exit
    sys.exit(code)



#
# RepoCommand.py
#
# the cati project
# Copyright 2020-2021 parsa shahmaleki <parsampsh@gmail.com>
#
# This file is part of cati.
#
# cati is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cati is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cati.  If not, see <https://www.gnu.org/licenses/>.
##################################################

""" Repo command """

import os
from cati.cmdline.BaseCommand import BaseCommand
from cati.cmdline import pr, ansi
from cati.repo.Repo import Repo
from cati.repo import Scanner
from cati.cmdline.components import ReposListErrorShower
from cati.frontend import Env, RootRequired

class RepoCommand(BaseCommand):
    """ Repo command """
    def help(self):
        """
        manage repositories

        Usage: cati repos [options]

        Options:
        -e|--edit: open repositories config file with editor
        -a|--add [new-repo]: add new repository
        --scan [directory]: scans packages inside in a directory and creates data files for repo
        -q|--quiet: quiet output

        Repo config structure:
        <url> pkg=<type(s) of packages. for example `cati` or `deb`> arch=<wanted architecture> channel=<wanted version channel(s)> name=<an name for repo> priority=<priority between another repos>
        
        Example:
        https://pkg.example.com/packages pkg=cati arch=all,i386 name=main-repo
        file:///path/to/packages pkg=cati,deb arch=all channel=alpha,security name=deb-repo
        """
        pass

    def config(self) -> dict:
        """ Define and config this command """
        return {
            'name': 'repo',
            'options': {
                '-q': [False, False],
                '--quiet': [False, False],
                '--edit': [False, False],
                '-e': [False, False],
                '--add': [False, False],
                '-a': [False, False],
                '--scan': [False, False],
            },
            'max_args_count': None,
            'min_args_count': None,
        }

    def run(self):
        """ Run command """

        if self.has_option('--edit') or self.has_option('-e'):
            return os.system('vim "' + Env.repos_config() + '"')

        if self.has_option('--add') or self.has_option('-a'):
            RootRequired.require_root_permission()
            repo_string = ''
            for arg in self.arguments:
                repo_string += arg + ' '
            repo_string = repo_string.strip()
            tmp_repo = Repo(repo_string)
            tmp_repo.loaded_from_file = 'argument'
            tmp_repo.line_number = 0
            if not tmp_repo.successful_loaded:
                ReposListErrorShower.show([tmp_repo])
                return 1
            # write repo
            path = Env.repos_config_dir('/' + tmp_repo.name + '-' + tmp_repo.get_pkg_str() + '-' + tmp_repo.get_arch_str())
            tmp = ''
            tmp_i = 1
            while os.path.isfile(path + tmp):
                tmp = '-' + str(tmp_i)
            f = open(path, 'w')
            f.write('# added manually\n' + repo_string)
            f.close()
            return 0

        if self.has_option('--scan'):
            for arg in self.arguments:
                if os.path.isdir(arg):
                    Scanner.scan(arg)
                    pr.p(ansi.green + 'directory ' + arg + ' scanned successfully' + ansi.reset)
                else:
                    self.message('directory' + arg + ' not found', is_error=True)
            return 0

        # show list of repos
        if not self.is_quiet():
            pr.p('Loading repositories list...')
        repos = Repo.get_list()
        if not self.is_quiet():
            pr.p('============================')
        ReposListErrorShower.show(repos)
        for repo in repos:
            if repo.successful_loaded:
                is_disable = ''
                if repo.is_disable:
                    is_disable = ' (Disabled)'
                pr.p(repo.name + ': ' + repo.url + ' pkg=' + repo.get_pkg_str() + ' arch=' + repo.get_arch_str() + ' channel=' + repo.get_channel_str() + is_disable)
